import java.io.DataInput;
import java.io.FileInputStream;
import java.io.DataInputStream;
import java.io.BufferedInputStream;
import java.io.OutputStreamWriter;
import java.io.IOException;
import java.io.EOFException;
import java.util.List;
import java.util.ArrayList;

/**
Unserializes serialized objects and writes them out in a readable, diffable text format.

@see <a href="http://java.sun.com/javase/6/docs/platform/serialization/spec/protocol.html">Object Serialization Stream Protocol</a>
*/
public class Unserialize {
	
    private static final int MAGIC = 0xaced;
    private static final int VERSION_2 = 5; // don't ask me, man, i didn't do it
    public static final int FIRST_HANDLE = 0x7e0000;
    
    private static final byte TC_NULL = 0x70;
    private static final byte TC_REFERENCE = 0x71;
    private static final byte TC_CLASSDESC = 0x72;
    private static final byte TC_OBJECT = 0x73;
    private static final byte TC_STRING = 0x74;
    private static final byte TC_ARRAY = 0x75;
    private static final byte TC_BLOCKDATA = 0x77;
    private static final byte TC_ENDBLOCKDATA = 0x78;
	/*
    private static final byte TC_CLASS = 0x76;
    private static final byte TC_RESET = 0x79;
    private static final byte TC_BLOCKDATALONG = 0x7a;
    private static final byte TC_EXCEPTION = 0x7b;
    private static final byte TC_LONGSTRING = 0x7c;
    private static final byte TC_PROXYCLASSDESC = 0x7d;
    private static finalU byte TC_ENUM = 0x7e;
	*/
    
    public static final byte FT_BYTE = (byte)'B';
    public static final byte FT_CHAR = (byte)'C';
    public static final byte FT_DOUBLE = (byte)'D';
    public static final byte FT_FLOAT = (byte)'F';
    public static final byte FT_INT = (byte)'I';
    public static final byte FT_LONG = (byte)'J';
    public static final byte FT_SHORT = (byte)'S';
    public static final byte FT_BOOLEAN = (byte)'Z';
    public static final byte FT_OBJECT = (byte)'L';
    public static final byte FT_ARRAY = (byte)'[';
	
	public static void main(String... args) throws IOException {
		Unserialize app = new Unserialize(new IndentingBufferedWriter(new OutputStreamWriter(System.out)));
		if (args.length == 0) app.unserialize(new DataInputStream(new BufferedInputStream(System.in)));
		else for (String arg: args) {
			app.unserialize(new DataInputStream(new BufferedInputStream(new FileInputStream(arg))));
			app.reset();
		}
	}
	
	private IndentingBufferedWriter out;
	private List<ContentDesc> objects;
	
	public Unserialize(IndentingBufferedWriter out) {
		this.out = out;
		reset();
	}
	
	public void reset() {
		objects = new ArrayList<ContentDesc>();
	}
	
	private void store(ContentDesc content) throws IOException {
		int handle = FIRST_HANDLE + objects.size();
		objects.add(content);
		content.setHandle(handle);
	}
	
	private ContentDesc retrieve(int handle) throws IOException {
		int idx = handle - FIRST_HANDLE;
		if (idx < 0) throw new IOException("bad handle: " + formatHandle(handle));
		if (idx >= objects.size()) throw new IOException("bad handle: " + formatHandle(handle));
		ContentDesc obj = objects.get(idx);
		return obj;
	}
	
	public void unserialize(DataInput in) throws IOException {
		readHeader(in);
		List<ContentDesc> contents = readContents(in);
		new ContentWriter(out).write(contents, objects);
	}
	
	private void readHeader(DataInput in) throws IOException {
		int magic = in.readUnsignedShort();
		if (magic != MAGIC) throw new IOException("bad magic: 0x" + Integer.toHexString(magic));
		int version = in.readUnsignedShort();
		if (version != VERSION_2) throw new IOException("bad version: 0x" + Integer.toHexString(magic));
	}
	
	private List<ContentDesc> readContents(DataInput in) throws IOException {
		List<ContentDesc> contents = new ArrayList<ContentDesc>();
		ContentDesc content;
		while ((content = readContent(in)) != null) contents.add(content);
		return contents;
	}
	
	private ContentDesc readContent(DataInput in) throws IOException {
		byte typeCode;
		try {
			typeCode = in.readByte();
		}
		catch (EOFException e) {
			return null;
		}
		switch (typeCode) {
			case TC_NULL: return null;
			case TC_REFERENCE: return readPrevObject(in);
			case TC_OBJECT: return readObject(in);
			case TC_STRING: return readString(in);
			case TC_ARRAY: return readArray(in);
			case TC_BLOCKDATA: return readBlockDataShort(in);
			case TC_ENDBLOCKDATA: return null; // only applies during an annotation read
			default: throw new IOException("bad content typecode: 0x" + Integer.toHexString(typeCode & 0xff));
		}
	}
	
	private ContentDesc readPrevObject(DataInput in) throws IOException {
		int handle = in.readInt();
		return retrieve(handle);
	}
	
	private ContentDesc readObject(DataInput in) throws IOException {
		// TC_OBJECT classDesc newHandle classdata[]  // data for each class
		// Data for each serializable class of the object, with the highest superclass first.
		ClassDesc classDesc = readClassDesc(in);
		List<ClassData> data = new ArrayList<ClassData>(classDesc.getDepth());
		ObjectDesc obj = new ObjectDesc(classDesc, data);
		store(obj);
		readAllClassData(in, classDesc, data);
		return obj;
	}
	
	private ClassDesc readClassDesc(DataInput in) throws IOException {
		// newClassDesc - TC_CLASSDESC TC_PROXYCLASSDESC
		// nullReference - TC_NULL
		// (ClassDesc)prevObject - TC_REFERENCE
		byte typeCode = in.readByte();
		switch (typeCode) {
			case TC_CLASSDESC: return readNewClassDesc(in);
			case TC_NULL: return null;
			case TC_REFERENCE: return (ClassDesc)readPrevObject(in);
			default: throw new IOException("bad classDesc typecode: 0x" + Integer.toHexString(typeCode & 0xff));
		}
	}
	
	private ClassDesc readNewClassDesc(DataInput in) throws IOException {
		// TC_CLASSDESC className serialVersionUID newHandle classDescInfo
		// classDescInfo:
		//  classDescFlags fields classAnnotation superClassDesc
		// fields:
		//  (short)<count>  fieldDesc[count]
		// classAnnotation:
		//  endBlockData = TC_ENDBLOCKDATA
		//  contents endBlockData      // contents written by annotateClass

		String className = in.readUTF();
		long serialVersionUID = in.readLong();
		byte flags = in.readByte();
		int numFields = in.readUnsignedShort();
		List<FieldDesc> fields = new ArrayList<FieldDesc>(numFields);
		ClassDesc classDesc = new ClassDesc(className, serialVersionUID, flags, fields);
		store(classDesc);
		for (int i = 0; i < numFields; ++i) fields.add(readFieldDesc(in));
		readAnnotation(in); // TODO: store this!
		ClassDesc superClassDesc = readClassDesc(in);
		if (superClassDesc != null) classDesc.setSuperClass(superClassDesc);
		return classDesc;
	}
	
	private FieldDesc readFieldDesc(DataInput in) throws IOException {
		// fieldDesc:
		//  primitiveDesc
		//  objectDesc
		// primitiveDesc:
		//  prim_typecode fieldName
		// objectDesc:
		//  obj_typecode fieldName className1
		byte fieldType = in.readByte();
		boolean isObject = FieldDesc.isObject(fieldType);
		String name = in.readUTF();
		StringDesc fieldClass;
		if (isObject) fieldClass = (StringDesc)readContent(in);
		else fieldClass = null;
		return new FieldDesc(fieldType, name, fieldClass);
	}
	
	private void readAllClassData(DataInput in, ClassDesc classDesc, List<ClassData> data) throws IOException {
		ClassDesc superClassDesc = classDesc.getSuperClass();
		if (superClassDesc != null) readAllClassData(in, superClassDesc, data);
		data.add(readClassData(in, classDesc));
	}
	
	private ClassData readClassData(DataInput in, ClassDesc classDesc) throws IOException {
		//  nowrclass                 // SC_SERIALIZABLE & classDescFlag && !(SC_WRITE_METHOD & classDescFlags)
		//  wrclass objectAnnotation  // SC_SERIALIZABLE & classDescFlag && SC_WRITE_METHOD & classDescFlags
		//  externalContents          // SC_EXTERNALIZABLE & classDescFlag && !(SC_BLOCKDATA  & classDescFlags
		//  objectAnnotation          // SC_EXTERNALIZABLE & classDescFlag&& SC_BLOCKDATA & classDescFlags
		// nowrclass:
		//  values                    // fields in order of class descriptor
		// wrclass:
		//  nowrclass
		// objectAnnotation:
		//  endBlockData
		//  contents endBlockData
		ClassData data;
		if (!classDesc.isExternalizable()) {
			data = readValues(in, classDesc);
		}
		else {
			if (!classDesc.isAnnotated()) throw new IOException("external contents not allowed in version 2 streams");
			data = new ClassData(classDesc);
		}
		if (classDesc.isAnnotated()) data.setAnnotation(readAnnotation(in));
		return data;
	}
	
	private ClassData readValues(DataInput in, ClassDesc classDesc) throws IOException {
		// The size and types are described by the
		// classDesc for the current object
		// docs are very vague so this is mostly made up
		ClassData data = new ClassData(classDesc);
		for (FieldDesc field: classDesc.getFields()) {
			Object value = readValue(in, field.getType());
			data.put(field, value);
		}
		return data;
	}
	
	private Object readValue(DataInput in, byte fieldType) throws IOException {
		switch (fieldType) {
			case FT_BYTE: return in.readByte();
			case FT_SHORT: return in.readShort();
			case FT_INT: return in.readInt();
			case FT_LONG: return in.readLong();
			case FT_CHAR: return in.readChar();
			case FT_FLOAT: return in.readFloat();
			case FT_DOUBLE: return in.readDouble();
			case FT_BOOLEAN: return in.readBoolean();
			case FT_OBJECT: return readContent(in); // TODO: check is an ObjectDesc or StringDesc
			case FT_ARRAY: return (ArrayDesc)readContent(in);
			default: throw new IOException("bad field type: 0x" + Integer.toHexString(fieldType & 0xff) + " (" + (char)fieldType + ")");
		}
	}
	
	private List<ContentDesc> readAnnotation(DataInput in) throws IOException {
		return readContents(in);
	}
	
	private StringDesc readString(DataInput in) throws IOException {
		// TC_STRING newHandle (utf)
		String str = in.readUTF();
		StringDesc strDesc = new StringDesc(str);
		store(strDesc);
		return strDesc;
	}
	
	private ArrayDesc readArray(DataInput in) throws IOException {
		// TC_ARRAY classDesc newHandle (int)<size> values[size]
		ClassDesc classDesc = readClassDesc(in);
		if (!classDesc.isArray()) throw new IOException("not an array: " + classDesc);
		int length = in.readInt();
		List<Object> values = new ArrayList<Object>(length);
		ArrayDesc array = new ArrayDesc(classDesc, values);
		store(array);
		byte fieldType = classDesc.getArrayElementFieldType();
		for (int i = 0; i < length; ++i) values.add(readValue(in, fieldType));
		return array;
	}
	
	private BlockDataDesc readBlockDataShort(DataInput in) throws IOException {
		//   TC_BLOCKDATA (unsigned byte)<size> (byte)[size]
		int len = in.readUnsignedByte();
		byte[] data = new byte[len];
		in.readFully(data);
		return new BlockDataDesc(data);
	}
	
	public static String formatHandle(int handle) {
		if (handle == -1) throw new IllegalArgumentException("void handle");
		return "@" + Integer.toHexString(handle);
	}
	
	public static String formatHandle(ContentDesc contentDesc) {
		return formatHandle(contentDesc.getHandle());
	}
	
}
